# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2023 Scipp contributors (https://github.com/scipp)

from ._disk_chopper_repr import disk_chopper_html_repr

__all__ = ['disk_chopper_html_repr']
